from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define


T = TypeVar("T", bound="VersionInfoResponse200")


@_attrs_define
class VersionInfoResponse200:
    """VersionInfoResponse200 model

    Attributes:
        scheme (str):
        server_name (str):
        version (str):
    """

    scheme: str
    server_name: str
    version: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        scheme = self.scheme
        server_name = self.server_name
        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "scheme": scheme,
                "server_name": server_name,
                "version": version,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`VersionInfoResponse200` from a dict"""
        d = src_dict.copy()
        scheme = d.pop("scheme")

        server_name = d.pop("server_name")

        version = d.pop("version")

        version_info_response_200 = cls(
            scheme=scheme,
            server_name=server_name,
            version=version,
        )

        return version_info_response_200



import json
import re
from collections.abc import Iterator
from typing import Any
from typing import cast
from typing import Generic
from typing import Literal
from typing import TypedDict
from typing import TypeGuard
from typing import TypeVar

import pydelfini
from pydelfini import exceptions
from pydelfini.collections import DelfiniFolder
from pydelfini.collections import DelfiniItem


class CycleMarker(TypedDict):
    n: int


def is_cycle_marker(val: Any) -> TypeGuard[CycleMarker]:
    return isinstance(val, dict) and "n" in val


T = TypeVar("T")


class CycleDetectQueue(Generic[T]):
    """Auto detect cycles in a queue.

    >>> q = CycleDetectQueue()
    >>> q.append('one')
    >>> q.append('two')
    >>> q.append('three')
    >>> _q = iter(q)
    >>> next(_q)
    'one'
    >>> next(_q)
    'two'
    >>> q.repeat('two')
    >>> next(_q)
    'three'
    >>> q.repeat('three')
    >>> next(_q)
    'two'
    >>> next(_q)
    'three'
    >>> q.repeat('three')
    >>> next(_q)
    'three'
    >>> q.repeat('three')
    >>> next(_q)
    Traceback (most recent call last):
        ...
    Exception: cycle detected, aborting

    """

    def __init__(self) -> None:
        self.items: list[T | CycleMarker] = [CycleMarker(n=0)]

    def append(self, item: T) -> None:
        marker = self.items.pop()
        assert is_cycle_marker(marker)
        self.items.append(item)
        marker["n"] = len(self.items)
        self.items.append(marker)

    def repeat(self, item: T) -> None:
        self.items.append(item)

    def __iter__(self) -> Iterator[T]:
        while self.items:
            item = self.items.pop(0)
            if is_cycle_marker(item):
                if item["n"] == 0:
                    return
                elif len(self.items) == item["n"]:
                    raise Exception("cycle detected, aborting")

                self.items.append(CycleMarker(n=len(self.items)))

            else:
                yield cast(T, item)


def do_copy(source_base: str, dest_base: str, collection_name: str) -> None:
    print("=== Login to source ===")
    source = pydelfini.login(source_base)
    if dest_base == source_base:
        dest = source
    else:
        print("=== Login to destination ===")
        dest = pydelfini.login(dest_base)

    source_collection = source.get_collection_by_name(collection_name)

    print()
    print("---- CREATING ----")

    print("[collection]", source_collection.name)
    dest_collection = dest.new_collection(
        source_collection.name,
        source_collection.description,
        source_collection.metadata,
    )

    map_q = CycleDetectQueue[DelfiniItem]()

    for item in source_collection.walk():
        if item.type == "folder":
            print("[folder]", item.path)
            base = (
                cast(DelfiniFolder, dest_collection[item.in_folder.path])
                if item.in_folder.path
                else dest_collection
            )
            base.new_folder(item.name)

        elif item.type == "dataview":
            map_q.append(item)

        elif item.type in ("file", "dictionary"):
            print(f"[{item.type}]", item.path)

            with dest_collection.open(
                item.path,
                "wb",
                type=cast(Literal["file", "dictionary"], item.type),
                parser=item.parser,
                columns=item.columns,
                metadata=item.metadata,
            ) as fp:
                fp.write(item.open("rb").read())

        elif item.type == "link":
            # TBD - need to have method for creating links
            pass

    # handle remapping dataviews
    for item in map_q:
        dv = item.open("r").read()
        dv = dv.replace("pigeon", "delfini")

        source_ids = re.findall(r"from `([-0-9a-f]+)`", dv)
        source_dest_map: dict[str, str] = {}
        for iid in source_ids:
            source_item = source_collection.get_item_by_id(iid)
            try:
                dest_item = dest_collection[source_item.path]
                source_dest_map[iid] = dest_item.id
            except exceptions.NotFoundError:
                print("Skipping dataview", item.path, "- will retry")
                pass

        if set(source_ids) != source_dest_map.keys():
            map_q.repeat(item)

        else:
            mapped_dv = dv
            for source_id, dest_id in source_dest_map.items():
                mapped_dv = mapped_dv.replace(source_id, dest_id)

            print("[dataview]", item.path)

            with dest_collection.open(
                item.path,
                "w",
                type="dataview",
                parser=item.parser,
                columns=item.columns,
                metadata=item.metadata,
            ) as fp:
                fp.write(mapped_dv)

    # update links and documents
    if "x-links-documents" in source_collection.metadata:
        links_docs = json.loads(source_collection.metadata["x-links-documents"])
        for doc in links_docs:
            if hit := re.match(r"delfini:([-0-9a-f]+)", doc.get("url", "")):
                source_item = source_collection.get_item_by_id(hit.group(1))
                print("[x-links-documents]", source_item.path)
                dest_item = dest_collection[source_item.path]
                doc["url"] = "delfini:" + dest_item.id

        new_metadata = source_collection.metadata.copy()
        new_metadata["x-links-documents"] = json.dumps(links_docs)

        dest_collection.set_metadata(new_metadata)


if __name__ == "__main__":
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("--source-base", required=True)
    argparser.add_argument("--dest-base", required=True)
    argparser.add_argument("collection_name")

    args = argparser.parse_args()

    do_copy(args.source_base, args.dest_base, args.collection_name)



from enum import Enum


class SystemConfigurationMotdAdditionalPropertyLevel(str, Enum):
    ALERT = "alert"
    INFO = "info"
    NOTICE = "notice"
    WARNING = "warning"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.motd import Motd


T = TypeVar("T", bound="RootMessageOfTheDayResponse200")


@_attrs_define
class RootMessageOfTheDayResponse200:
    """RootMessageOfTheDayResponse200 model

    Attributes:
        messages (List['Motd']):
    """

    messages: List["Motd"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        messages = []
        for messages_item_data in self.messages:
            messages_item = messages_item_data.to_dict()
            messages.append(messages_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "messages": messages,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`RootMessageOfTheDayResponse200` from a dict"""
        d = src_dict.copy()
        messages = []
        _messages = d.pop("messages")
        for messages_item_data in _messages:
            messages_item = Motd.from_dict(messages_item_data)

            messages.append(messages_item)

        root_message_of_the_day_response_200 = cls(
            messages=messages,
        )

        return root_message_of_the_day_response_200



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.motd_level import MotdLevel


T = TypeVar("T", bound="Motd")


@_attrs_define
class Motd:
    """Message of the day

    Attributes:
        level (MotdLevel):
        message (str):
    """

    level: MotdLevel
    message: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        level = self.level.value
        message = self.message

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "level": level,
                "message": message,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`Motd` from a dict"""
        d = src_dict.copy()
        level = MotdLevel(d.pop("level"))

        message = d.pop("message")

        motd = cls(
            level=level,
            message=message,
        )

        motd.additional_properties = d
        return motd

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os.path
import re
from typing import cast
from typing import Optional

import boto3
import pydelfini
from botocore import UNSIGNED
from botocore.config import Config
from pydelfini.collections import FolderMixin
from pydelfini.exceptions import NotFoundError
from tqdm import tqdm

s3 = boto3.resource("s3", config=Config(signature_version=UNSIGNED))


class S3URL:
    """Parses an S3 URL.

    >>> u = S3URL('s3://example/path')
    >>> u.bucket
    'example'
    >>> u.path
    'path'

    >>> v = S3URL('s3://example/')
    >>> v.bucket
    'example'
    >>> v.path
    ''

    Raises ValueError if not parsable.

    >>> S3URL('something else')
    Traceback (most recent call last):
      ...
    ValueError: expect S3 URL of format s3://bucket-name/optional-path

    """

    def __init__(self, url: str):
        result = re.match(r"s3://([^/]+)/(.*)", url)
        if not result:
            raise ValueError("expect S3 URL of format s3://bucket-name/optional-path")
        self.bucket = result.group(1)
        self.path = result.group(2)

    def __str__(self) -> str:
        return f"s3://{self.bucket}/{self.path}"


def mkdirs(base: FolderMixin, path: str) -> FolderMixin:
    if path.startswith("/"):
        path = path.removeprefix("/")

    try:
        return cast(FolderMixin, base[path])
    except NotFoundError:
        dirs, name = os.path.split(path)
        if dirs:
            folder = mkdirs(base, dirs)
            return folder.new_folder(name)
        else:
            return base.new_folder(name)


def do_link(
    srv_base: str,
    scheme: str,
    collection_name: str,
    s3_url: S3URL,
    in_folder: Optional[str],
) -> None:
    client = pydelfini.login(srv_base, scheme=scheme)
    collection = client.get_collection_by_name(collection_name)
    if in_folder:
        base_folder = mkdirs(collection, in_folder)
    else:
        base_folder = collection

    bucket = s3.Bucket(s3_url.bucket)
    for obj in tqdm(bucket.objects.filter(Prefix=s3_url.path)):
        obj_path = obj.key.removeprefix(s3_url.path)
        obj_dir, obj_name = os.path.split(obj_path)
        if not obj_name:
            continue

        if obj_dir:
            folder = mkdirs(base_folder, obj_dir)
        else:
            folder = base_folder

        link_target = f"s3://{obj.bucket_name}/{obj.key}"
        folder.new_link(link_target, obj_name)


if __name__ == "__main__":
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--delfini-base",
        "-H",
        required=True,
        metavar="HOSTNAME",
        help="connect to HOSTNAME",
    )
    argparser.add_argument(
        "--insecure",
        dest="scheme",
        default="https",
        action="store_const",
        const="http",
        help="connect using http rather than https",
    )
    argparser.add_argument("--path", "-f")
    argparser.add_argument("s3_url", type=S3URL)
    argparser.add_argument("collection_name")

    args = argparser.parse_args()

    do_link(
        args.delfini_base, args.scheme, args.collection_name, args.s3_url, args.path
    )



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="AdminSetConfigTogglesBody")


@_attrs_define
class AdminSetConfigTogglesBody:
    """AdminSetConfigTogglesBody model

    Attributes:
        unique_user_email (Union[Unset, bool]):
    """

    unique_user_email: Union[Unset, bool] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        unique_user_email = self.unique_user_email

        field_dict: Dict[str, Any] = {}
        field_dict.update({})
        if unique_user_email is not UNSET:
            field_dict["unique_user_email"] = unique_user_email

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`AdminSetConfigTogglesBody` from a dict"""
        d = src_dict.copy()
        unique_user_email = d.pop("unique_user_email", UNSET)

        admin_set_config_toggles_body = cls(
            unique_user_email=unique_user_email,
        )

        return admin_set_config_toggles_body



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="SessionMembershipsMembershipsItem")


@_attrs_define
class SessionMembershipsMembershipsItem:
    """SessionMembershipsMembershipsItem model

    Attributes:
        id (str):
        type (str):
        role (Union[Unset, str]):
    """

    id: str
    type: str
    role: Union[Unset, str] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        id = self.id
        type = self.type
        role = self.role

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "id": id,
                "type": type,
            }
        )
        if role is not UNSET:
            field_dict["role"] = role

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SessionMembershipsMembershipsItem` from a dict"""
        d = src_dict.copy()
        id = d.pop("id")

        type = d.pop("type")

        role = d.pop("role", UNSET)

        session_memberships_memberships_item = cls(
            id=id,
            type=type,
            role=role,
        )

        return session_memberships_memberships_item



import argparse
import ast
import os
import random
import re
import shlex
import time
from pathlib import Path
from urllib.request import urlopen

import pytest
from httpx import RemoteProtocolError
from pydelfini.client import make_base_url
from pydelfini.dcmd import cli
from pydelfini.dcmd.commands import Commands as PcmdCommands
from pydelfini.delfini_core import Client as CoreClient
from pydelfini.delfini_core.api.user import user_create_user
from pydelfini.delfini_core.login import Login
from pydelfini.delfini_core.login import LoginError
from pydelfini.delfini_core.login import to_token_file
from pydelfini.delfini_core.models import NewUser
from yaml import dump as yaml_dump
from yaml import safe_load


# The tests in this test suite are designed to be run against a local
# development instance of Delfini hosted at localhost:8170, similar to
# the Cypress tests in `delfini-nuxt`. In order to be able to run these
# tests, be sure to have the local development instance running in the
# background with the command:
#
#    flask run --debug
#


@pytest.fixture
def local_user(tmp_path):
    # try to log in with the credentials
    username, password = "dcmd.tester", "insecure"
    base_url = make_base_url("localhost:8170", "http")
    try:
        authenticated = Login(CoreClient(base_url=base_url)).with_password(
            username, password, sudo=True
        )
    except LoginError:
        print("Login failed, trying to create new user...")
        user_create_user.sync(
            body=NewUser(
                given_name="dcmd",
                last_name="tester",
                user_email="john.doe@example.com",
                user_name=username,
                password=password,
            ),
            client=CoreClient(base_url=base_url),
        )
        authenticated = Login(CoreClient(base_url=base_url)).with_password(
            username, password, sudo=True
        )

    token_file = tmp_path / "token"
    to_token_file(authenticated, token_file)

    yield str(token_file)


def random_id():
    return f"rr{random.randint(0, 1048576):05x}rr"


def mask(rv):
    # mask UUIDs
    rv = re.sub(
        r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
        "00000000-0000-0000-0000-000000000000",
        rv,
    )
    # mask timestamps
    rv = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.[0-9+:]*)?",
        "1970-01-02T03:04:05.000000",
        rv,
    )
    # mask task worker
    rv = re.sub(r"worker: .*", "worker: dcmd-test-mask", rv)
    # mask random ID
    rv = re.sub(r"rr[0-9a-f]{5}rr", "random-id", rv)
    # mask bundle ID
    rv = re.sub(r"b-[a-zA-Z0-9]{8}", "b-00000000", rv)

    return rv


@pytest.fixture
def dcmd(local_user, capsys):
    def _cmd(args, masked=True):
        if isinstance(args, str):
            args = shlex.split(args)

        try:
            cli.main(["-H", "localhost:8170", "--insecure", "-k", local_user] + args)
        except RemoteProtocolError:
            time.sleep(1)
            return _cmd(args, masked=masked)

        outv = capsys.readouterr()
        if not outv.err:
            rv = outv.out
            if masked:
                rv = mask(rv)
            return rv

        return outv

    yield _cmd


@pytest.fixture
def test_target_user(dcmd):
    response = user_create_user.sync(
        body=NewUser(
            given_name="dcmd",
            last_name="test target",
            user_email=f"{random_id()}@example.com",
            user_name="dcmd.test.target",
            password=random_id(),
        ),
        client=CoreClient(base_url=make_base_url("localhost:8170", "http")),
    )

    yield response.subject_id

    dcmd("users delete dcmd.test.target")


@pytest.fixture
def scratch_collection(dcmd):
    collection_name = f"dcmd-test.{random_id()}"
    rv = safe_load(dcmd(f"collection new {collection_name}", masked=False))

    yield collection_name, rv["id"]

    dcmd(f"collection delete {rv['id']}")


def test_dcmd_auth_whoami(dcmd, snapshot):
    dcmd('auth set-metadata x-tos \'{"toc":3,"coc":6,"priv":3,"com":1}\'')

    assert dcmd("auth whoami") == snapshot


# NOTE: If the following tests are failing with "Unexpected status
# code: 403", you should grant the `dcmd.tester` user full admin
# access to your local development instance with the command:
#
#   flask admin grant dcmd.tester
#


def test_dcmd_tasks(dcmd, snapshot):
    stats = safe_load(dcmd("tasks stats"))
    assert stats.keys() == {"queued", "results", "schedules"}

    schedules = safe_load(dcmd("tasks schedules"))["schedules"]
    assert len(schedules) >= 2

    results = safe_load(dcmd("tasks results --idemkey tasks-backends-sqlite-vacuum"))[
        "results"
    ]
    if len(results) > 0:
        assert results[0] == snapshot


def test_dcmd_users(dcmd, test_target_user):
    rv = dcmd("users list")
    assert set(rv.splitlines()[0].split()) == {
        "account_id",
        "created_on",
        "metadata",
        "subject_id",
        "user_name",
        "visibility_level",
        "given_name",
        "groups",
        "last_name",
        "user_email",
    }
    assert "dcmd.test.target" in rv

    # disable and enable
    rv = safe_load(dcmd("users disable dcmd.test.target"))
    assert rv["disabled_on"] is not None

    rv = safe_load(dcmd("users enable dcmd.test.target"))
    assert rv["disabled_on"] is None


def test_dcmd_groups(dcmd, test_target_user, snapshot):
    group_name = f"test-dcmd-group.{random_id()}"
    new_group = dcmd(f"groups new {group_name}", masked=False)
    group_id = safe_load(new_group)["id"]
    assert mask(new_group) == snapshot

    groups = safe_load(dcmd("groups list"))["groups"]
    assert safe_load(mask(new_group)) in groups

    dcmd(f"groups join {group_id} {test_target_user}")

    assert len(safe_load(dcmd(f"groups list-members {group_id}"))["members"]) == 2

    dcmd(f"groups update --name new-test-group {group_id}")

    dcmd(f"groups delete {group_id}")


def test_dcmd_accounts(dcmd, snapshot, tmp_path, scratch_collection):
    collection_name, collection_id = scratch_collection

    account_name = f"test-dcmd-account.{random_id()}"
    new_account = dcmd(f"accounts new {account_name} --description heyyo", masked=False)
    account_id = safe_load(new_account)["id"]
    assert mask(new_account) == snapshot

    accounts = safe_load(dcmd("accounts list", masked=False))["accounts"]
    assert safe_load(new_account) in accounts

    # update space
    index_file = tmp_path / "index.md"
    with open(index_file, "w") as fp:
        fp.write("# Hello world!\n")
    dcmd(f"collection upload {collection_name} {index_file}")
    dcmd(f"accounts update-space {account_id} {collection_id}")
    response = urlopen(
        f"http://localhost:8170/api/v1/account/{account_id}/pages/index.md"
    )
    assert response.code == 200

    dcmd(f"accounts delete {account_id}")

    accounts = safe_load(dcmd("accounts list", masked=False))["accounts"]
    assert safe_load(new_account) not in accounts


def test_dcmd_admin_grant(dcmd, test_target_user):
    dcmd(f"admin grant-admin -o group {test_target_user}")

    rv = dcmd("admin list-admins", masked=False)
    assert test_target_user in rv

    dcmd(f"admin revoke-admin {test_target_user}")

    rv = dcmd("admin list-admins", masked=False)
    assert test_target_user not in rv


def test_dcmd_collection(dcmd, tmp_path, scratch_collection):
    collection_name, collection_id = scratch_collection
    rv = dcmd("collection list", masked=False)
    assert collection_name in rv
    assert collection_id in rv

    test_file = tmp_path / "test.txt"
    with open(test_file, "w") as fp:
        fp.write("hello world this is a test\n")

    dcmd(f"collection upload {collection_name} {test_file}")

    test_file.unlink()

    old_cwd = os.getcwd()
    os.chdir(tmp_path)
    dcmd(f"collection download {collection_name} test.txt")
    os.chdir(old_cwd)

    assert test_file.exists()


def test_dcmd_pdd(dcmd, scratch_collection, snapshot):
    collection_name, collection_id = scratch_collection
    common_pdd_fn = Path(__file__).parent / "common_pdd.json"

    rv = dcmd(f"pdd upload {collection_id} {common_pdd_fn}", masked=False)
    assert mask(rv) == snapshot
    safe_load(rv)["id"]

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot

    dcmd(f"pdd set-order {collection_name} common_pdd.json 6fr70ICV O8zPEsiM")

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot

    assert (
        dcmd(f"pdd bundle {collection_name} common_pdd.json MyBundle O8zPEsiM 6fr70ICV")
        == snapshot
    )

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot


def test_dcmd_cdeset(dcmd, scratch_collection):
    collection_name, collection_id = scratch_collection
    common_pdd_fn = Path(__file__).parent / "common_pdd.json"

    rv = dcmd(f"pdd upload {collection_id} {common_pdd_fn}", masked=False)
    item_id = safe_load(rv)["id"]

    cdeset_name = f"dcmd.{random_id()}"
    dcmd(f"cdeset new {cdeset_name} 'dcmd test cde set'")

    rv = dcmd("cdeset list", masked=False)
    assert cdeset_name in rv

    dcmd(f"cdeset copy-from-pdd {cdeset_name} {collection_id} LIVE {item_id}")

    rv = safe_load(dcmd("cdeset list", masked=False))
    cdeset = {"f": c for c in rv["cdesets"] if c["name"] == cdeset_name}.get("f")
    assert cdeset["num_cdes"] == 2
    assert cdeset["description"] == "dcmd test cde set"

    dcmd(f"cdeset delete {cdeset_name}")


def test_dcmd_admin_config_toggle(dcmd):
    rv = safe_load(dcmd("admin config toggle get --details"))
    assert "unique_user_email" in rv

    if not rv["details"]["unique_user_email"]["duplicates"]:
        dcmd("admin config toggle enable unique_user_email")
        dcmd("admin config toggle disable unique_user_email")


# This test needs to be at/near the bottom in order to minimize
# spurious test fails caused by triggering the backend reload due to
# config updates
def test_dcmd_admin_config(dcmd, tmp_path, capsys):
    rv = safe_load(dcmd("admin config get"))
    assert "datastores" in rv
    assert "authentication" in rv

    dcmd(
        "admin config set datastores.dcmd_test"
        + f' \'{{"enabled": false, "path": "{tmp_path}", "type": "local"}}\''
    )

    rv = safe_load(dcmd("admin config get", masked=False))
    assert "dcmd_test" in rv["datastores"]

    newdir = tmp_path / "new"
    rv["datastores"]["dcmd_test"]["path"] = str(newdir)

    config_fn = tmp_path / "config.yaml"
    with open(config_fn, "w") as fp:
        yaml_dump(rv, fp)

    dcmd(f"admin config put {config_fn}")

    rv = safe_load(dcmd("admin config get"))
    assert "dcmd_test" in rv["datastores"]
    assert rv["datastores"]["dcmd_test"]["path"] == str(newdir)

    dcmd("admin config delete datastores.dcmd_test")

    rv = safe_load(dcmd("admin config get"))
    assert "dcmd_test" not in rv["datastores"]


###
### Coverage Verification
###


class CoverageScanner(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.actions = {}

    def register_actions(self, parser, actlevel=None):
        if actlevel is None:
            actlevel = self.actions

        actions = parser._get_positional_actions()
        if actions and actions[0].choices:
            for arg, subparser in actions[0].choices.items():
                self.register_actions(subparser, actlevel.setdefault(arg, {}))

    def coverage_needed(self):
        needed = set()

        def _walk(d, s=None):
            if s is None:
                s = []
            if "SEEN" in d:
                return
            if not d:
                needed.add(" ".join(s))
            for k in d:
                _walk(d[k], s + [k])

        _walk(self.actions)
        return needed

    def visit_Call(self, node):
        if not isinstance(node.func, ast.Name):
            return self.generic_visit(node)
        if node.func.id != "dcmd":
            return self.generic_visit(node)
        if not node.args:
            return self.generic_visit(node)

        def _resolve(arg):
            if isinstance(arg, ast.Constant):
                return arg.value
            elif isinstance(arg, ast.JoinedStr):
                return _resolve(arg.values[0])
            elif isinstance(arg, ast.BinOp):
                return _resolve(arg.left)
            else:
                # The scanner is looking for `dcmd("....")` calls in
                # the code above. For each call, it splits the string
                # using shlex.split and uses that to determine which
                # command is being invoked. If you invoke dcmd using a
                # non-constant string, it can't use this simplistic
                # parsing method and will raise the exception below.
                # If you need to add parameters to your dcmd call, you
                # can use an f-string or plus-based string
                # concatenation, but in both cases, the left-most
                # substring must be constant.
                print(ast.dump(node, indent=2))
                raise Exception(
                    f"unexpected dcmd call on line {node.lineno}: {ast.unparse(node)}"
                )

        dcmd_arg = _resolve(node.args[0])

        actlevel = self.actions
        for word in shlex.split(dcmd_arg):
            if word in actlevel:
                actlevel = actlevel[word]

        if not actlevel:
            actlevel["SEEN"] = True


def test_dcmd_coverage():
    # This checks all of the dcmd() calls in the AST of this file and
    # matches them up with commands defined in the dcmd command set.
    # Any commands not called will show up in
    # `scanner.coverage_needed()`.

    argparser = argparse.ArgumentParser()
    PcmdCommands._add_to_parser(
        argparser.add_subparsers(title="commands", required=True)
    )
    scanner = CoverageScanner()
    scanner.register_actions(argparser)

    test_ast = ast.parse(open(__file__).read(), filename=__file__)
    scanner.visit(test_ast)

    assert scanner.coverage_needed() == set()



"""Update metadata on the collection"""

from http import HTTPStatus
from typing import Any
from typing import cast
from typing import Dict
from typing import Union

import httpx

from ... import errors
from ...client import AuthenticatedClient
from ...client import Client
from ...models.server_error import ServerError
from ...models.update_collection import UpdateCollection
from ...types import Response


def _get_kwargs(
    collection_id: str,
    version_id: str,
    *,
    body: UpdateCollection,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    _kwargs: Dict[str, Any] = {
        "method": "patch",
        "url": "/collections/{collection_id}/{version_id}".format(
            collection_id=collection_id,
            version_id=version_id,
        ),
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[Any, ServerError]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, {"attribute": "None", "return_type": "None"})
        return response_204
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ServerError.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ServerError.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = ServerError.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE:
        response_415 = ServerError.from_dict(response.json())

        return response_415
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = ServerError.from_dict(response.json())

        return response_500

    raise errors.UnexpectedStatus(response.status_code, response.content)


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, ServerError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: UpdateCollection,
) -> Response[Union[Any, ServerError]]:
    """Update metadata on the collection

    Update the collection's metadata, leaving unspecified fields as-is.

    This method currently operates the same as the corresponding
    PUT method. It may be deprecated in the future, or implemented
    according to the JSON Patch standard.

    Args:
        collection_id (str):
        version_id (str):
        body (UpdateCollection): Update core collection properties

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, ServerError]]
    """

    kwargs = _get_kwargs(
        collection_id=collection_id,
        version_id=version_id,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: UpdateCollection,
) -> Union[Any]:
    """Update metadata on the collection

    Update the collection's metadata, leaving unspecified fields as-is.

    This method currently operates the same as the corresponding
    PUT method. It may be deprecated in the future, or implemented
    according to the JSON Patch standard.

    Args:
        collection_id (str):
        version_id (str):
        body (UpdateCollection): Update core collection properties

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any]
    """

    response = sync_detailed(
        collection_id=collection_id,
        version_id=version_id,
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed


async def asyncio_detailed(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: UpdateCollection,
) -> Response[Union[Any, ServerError]]:
    """Update metadata on the collection

    Update the collection's metadata, leaving unspecified fields as-is.

    This method currently operates the same as the corresponding
    PUT method. It may be deprecated in the future, or implemented
    according to the JSON Patch standard.

    Args:
        collection_id (str):
        version_id (str):
        body (UpdateCollection): Update core collection properties

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, ServerError]]
    """

    kwargs = _get_kwargs(
        collection_id=collection_id,
        version_id=version_id,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: UpdateCollection,
) -> Union[Any]:
    """Update metadata on the collection

    Update the collection's metadata, leaving unspecified fields as-is.

    This method currently operates the same as the corresponding
    PUT method. It may be deprecated in the future, or implemented
    according to the JSON Patch standard.

    Args:
        collection_id (str):
        version_id (str):
        body (UpdateCollection): Update core collection properties

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any]
    """

    response = await asyncio_detailed(
        collection_id=collection_id,
        version_id=version_id,
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.session_memberships_memberships_item import (
    SessionMembershipsMembershipsItem,
)


T = TypeVar("T", bound="SessionMemberships")


@_attrs_define
class SessionMemberships:
    """SessionMemberships model

    Attributes:
        memberships (List['SessionMembershipsMembershipsItem']):
    """

    memberships: List["SessionMembershipsMembershipsItem"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        memberships = []
        for memberships_item_data in self.memberships:
            memberships_item = memberships_item_data.to_dict()
            memberships.append(memberships_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "memberships": memberships,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SessionMemberships` from a dict"""
        d = src_dict.copy()
        memberships = []
        _memberships = d.pop("memberships")
        for memberships_item_data in _memberships:
            memberships_item = SessionMembershipsMembershipsItem.from_dict(
                memberships_item_data
            )

            memberships.append(memberships_item)

        session_memberships = cls(
            memberships=memberships,
        )

        return session_memberships



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.config_toggles_details import ConfigTogglesDetails
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="ConfigToggles")


@_attrs_define
class ConfigToggles:
    """ConfigToggles model

    Attributes:
        unique_user_email (bool):
        details (Union[Unset, ConfigTogglesDetails]):
    """

    unique_user_email: bool
    details: Union[Unset, "ConfigTogglesDetails"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        unique_user_email = self.unique_user_email
        details: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.details, Unset):
            details = self.details.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "unique_user_email": unique_user_email,
            }
        )
        if details is not UNSET:
            field_dict["details"] = details

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`ConfigToggles` from a dict"""
        d = src_dict.copy()
        unique_user_email = d.pop("unique_user_email")

        _details = d.pop("details", UNSET)
        details: Union[Unset, ConfigTogglesDetails]
        if isinstance(_details, Unset):
            details = UNSET
        else:
            details = ConfigTogglesDetails.from_dict(_details)

        config_toggles = cls(
            unique_user_email=unique_user_email,
            details=details,
        )

        return config_toggles



import datetime
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define
from dateutil.parser import isoparse

from ..models.task_def import TaskDef
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="TaskQueueEntry")


@_attrs_define
class TaskQueueEntry:
    """TaskQueueEntry model

    Attributes:
        submitted_on (datetime.datetime):
        task (TaskDef):
        claimed (Union[Unset, str]):
        claimed_on (Union[Unset, datetime.datetime]):
        timeout_on (Union[Unset, datetime.datetime]):
    """

    submitted_on: datetime.datetime
    task: "TaskDef"
    claimed: Union[Unset, str] = UNSET
    claimed_on: Union[Unset, datetime.datetime] = UNSET
    timeout_on: Union[Unset, datetime.datetime] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        submitted_on = self.submitted_on.isoformat()
        task = self.task.to_dict()
        claimed = self.claimed
        claimed_on: Union[Unset, str] = UNSET
        if not isinstance(self.claimed_on, Unset):
            claimed_on = self.claimed_on.isoformat()
        timeout_on: Union[Unset, str] = UNSET
        if not isinstance(self.timeout_on, Unset):
            timeout_on = self.timeout_on.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "submitted_on": submitted_on,
                "task": task,
            }
        )
        if claimed is not UNSET:
            field_dict["claimed"] = claimed
        if claimed_on is not UNSET:
            field_dict["claimed_on"] = claimed_on
        if timeout_on is not UNSET:
            field_dict["timeout_on"] = timeout_on

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`TaskQueueEntry` from a dict"""
        d = src_dict.copy()
        submitted_on = isoparse(d.pop("submitted_on"))

        task = TaskDef.from_dict(d.pop("task"))

        claimed = d.pop("claimed", UNSET)

        _claimed_on = d.pop("claimed_on", UNSET)
        claimed_on: Union[Unset, datetime.datetime]
        if isinstance(_claimed_on, Unset):
            claimed_on = UNSET
        else:
            claimed_on = isoparse(_claimed_on)

        _timeout_on = d.pop("timeout_on", UNSET)
        timeout_on: Union[Unset, datetime.datetime]
        if isinstance(_timeout_on, Unset):
            timeout_on = UNSET
        else:
            timeout_on = isoparse(_timeout_on)

        task_queue_entry = cls(
            submitted_on=submitted_on,
            task=task,
            claimed=claimed,
            claimed_on=claimed_on,
            timeout_on=timeout_on,
        )

        return task_queue_entry



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.task_queue_entry import TaskQueueEntry


T = TypeVar("T", bound="TasksQueued")


@_attrs_define
class TasksQueued:
    """TasksQueued model

    Attributes:
        queue (List['TaskQueueEntry']):
    """

    queue: List["TaskQueueEntry"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        queue = []
        for queue_item_data in self.queue:
            queue_item = queue_item_data.to_dict()
            queue.append(queue_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "queue": queue,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`TasksQueued` from a dict"""
        d = src_dict.copy()
        queue = []
        _queue = d.pop("queue")
        for queue_item_data in _queue:
            queue_item = TaskQueueEntry.from_dict(queue_item_data)

            queue.append(queue_item)

        tasks_queued = cls(
            queue=queue,
        )

        return tasks_queued
